def load_dataset(path):
    buf = []
    for line in open(path, encoding="utf-8"):
        line = line.rstrip("\r\n")
        if not line:
            yield "".join(buf), buf
            buf = []
        else:
            buf.append(line)
    if buf:
        yield "".join(buf), buf


def evaluate_split(gold, pred):
    import re
    from difflib import SequenceMatcher

    ws = re.compile(r"\s+")
    full_str = "".join(ws.sub("", g) for g in gold)
    gold_spans = []
    pred_spans = []
    final_score = [0.0 for _ in gold]
    em = [0 for _ in gold]
    em_pred = [0 for _ in pred]
    pred_matches = [None for _ in pred]
    matches = []
    pred_ii = [[] for _ in pred]

    offset = 0
    for g in gold:
        l = len(ws.sub("", g))
        gold_spans.append((offset, offset + l))
        offset += l

    offset = 0
    for p in pred:
        p = ws.sub("", p)
        try:
            begin = full_str.index(p, offset)
            end = begin + len(p)
        except ValueError:
            diff = SequenceMatcher(a=full_str[offset:], b=p).get_opcodes()
            diff_start = next(i for i, (op, *_) in enumerate(diff) if op != "delete")
            diff_end = (
                -next(i for i, (op, *_) in enumerate(diff[::-1]) if op != "delete") - 1
            )
            begin = offset + diff[diff_start][1]
            end = offset + diff[diff_end][2]
        pred_spans.append((begin, end))
        offset = end

    start = 0
    for p, (pbegin, pend) in enumerate(pred_spans):
        matched = False
        for i in range(start, len(gold)):
            gbegin, gend = gold_spans[i]
            if pbegin == gbegin and pend == gend:
                final_score[i] = 1.0
                em[i] = 1
                em_pred[p] = 1
                pred_matches[p] = i
                break
            ibegin = max(pbegin, gbegin)
            iend = min(pend, gend)
            if ibegin < iend:
                r = 2 * (iend - ibegin) / ((gend - gbegin) + (pend - pbegin))
                matches.append((r, i, p))
                pred_ii[p].append(len(matches) - 1)
                matched = True
            elif matched:
                break
            else:
                start += 1

    while True:
        try:
            s, i, p = max(matches)
        except:
            break
        if final_score[i] < s:
            final_score[i] = s
            pred_matches[p] = i
        for n in pred_ii[p]:
            matches[n] = (0,)
    return final_score, em, em_pred, pred_matches


def run_evaluate(dataset, split_func, result_output=None, err_output=None):
    import time

    f1, norm_f1, em = [], [], []
    system_sents = 0
    elapsed = 0

    if result_output is not None:
        rout = open(result_output, "w", encoding="utf-8")
    if err_output is not None:
        fout = open(err_output, "w", encoding="utf-8")

    for text, gold in load_dataset(dataset):
        pcount = time.perf_counter()
        pred = split_func(text)
        elapsed += time.perf_counter() - pcount
        s, e, ep, pred_matches = evaluate_split(gold, pred)
        length_penalty = min(len(gold) / len(pred), 1)
        norm_s = [_s * length_penalty for _s in s]
        if result_output is not None:
            rout.write("\n".join(pred) + "\n\n")
            rout.flush()
        if err_output is not None:
            for exact, sent, matched_gold in zip(ep, pred, pred_matches):
                if exact:
                    continue
                r_score = None if matched_gold is None else s[matched_gold]
                n_score = None if matched_gold is None else norm_s[matched_gold]
                m_gold = None if matched_gold is None else gold[matched_gold]
                fout.write(f"{r_score}\t{n_score}\t{sent}\t{m_gold}\n")
            fout.write("\n")
            fout.flush()
        f1 += s
        norm_f1 += norm_s
        em += e
        system_sents += len(ep)

    if result_output is not None:
        rout.close()
    if err_output is not None:
        fout.close()

    print("[Sentence Split Benchmark] Dataset: " + dataset)
    print(
        f"Gold: {len(f1)} sents, "
        f"System: {system_sents} sents, "
        f"EM: {sum(em) / len(em):.5f}, "
        f"F1: {sum(f1) / len(f1):.5f}, "
        f"Normalized F1: {sum(norm_f1) / len(norm_f1):.5f}, "
        f"Latency: {elapsed*1000:.2f} msec"
    )
    print()



# This code was copied from g2pk [https://github.com/kyubyong/g2pK]
# And modified by Hyunwoong Ko [https://github.com/hyunwoongko]

import re
from functools import partial
from typing import Union, List, Tuple

from kss._modules.g2p.english import convert_eng
from kss._modules.g2p.numerals import convert_num
from kss._modules.g2p.regular import link1, link2, link3, link4
from kss._modules.g2p.special import (
    jyeo,
    ye,
    consonant_ui,
    josa_ui,
    vowel_ui,
    jamo,
    rieulgiyeok,
    rieulbieub,
    verb_nieun,
    balb,
    palatalize,
    modifying_rieul
)
from kss._modules.g2p.utils import (
    rule_id2text,
    annotate,
    gloss,
    group,
    table,
    convert_idioms,
)
from kss._modules.jamo._jamo import h2j, j2h
from kss._utils.multiprocessing import _run_job
from kss._utils.sanity_checks import _check_text, _check_num_workers, _check_type, _check_backend_mecab_pecab_only


def g2p(
    text: Union[str, List[str], Tuple[str]],
    descriptive: bool = False,
    group_vowels: bool = False,
    to_syllable: bool = True,
    convert_english_to_hangul_phonemes: bool = True,
    convert_numbers_to_hangul_phonemes: bool = True,
    num_workers: Union[int, str] = "auto",
    backend: str = "auto",
    verbose: bool = False,
) -> Union[str, List[str]]:
    """
    This function provides a way to convert Korean graphemes to phonemes.
    The 'grapheme' means a letter or a character, and the 'phoneme' means a sound.

    Args:
        text (Union[str, List[str], Tuple[str]]): single text or list of texts
        descriptive (bool): return descriptive pronunciation, the 'descriptive' means a real-life pronunciation
        group_vowels (bool): If True, the vowels of the identical sound are normalized. (e.g. ㅒ -> ㅖ)
        to_syllable (bool): If True, hangul letters or jamo are assembled to form syllables.
        convert_english_to_hangul_phonemes (bool): If True, convert English to Hangul phonemes
        convert_numbers_to_hangul_phonemes (bool): If True, convert numbers to Hangul phonemes
        num_workers (Union[int, str]): the number of multiprocessing workers
        backend (str): morpheme analyzer backend. 'mecab', 'pecab' are supported
        verbose (bool): whether to print verbose outputs or not

    Returns:
        Union[str, List[str]]: phoneme string or list of phoneme strings from the given text

    Examples:
        >>> from kss import Kss
        >>> g2p = Kss("g2p")
        >>> text = "어제는 맑았는데 오늘은 흐리다."
        >>> output = g2p(text)
        >>> print(output)
        "어제는 말간는데 오느른 흐리다."

    References:
        This was copied from [g2pk](https://github.com/Kyubyong/g2pk) and modified by Kss
    """
    text, finish = _check_text(text)

    if finish:
        return text

    descriptive = _check_type(descriptive, "descriptive", bool)
    group_vowels = _check_type(group_vowels, "group_vowels", bool)
    to_syllable = _check_type(to_syllable, "to_syllable", bool)
    convert_english_to_hangul_phonemes = _check_type(convert_english_to_hangul_phonemes,
                                                     "convert_english_to_hangul_phonemes", bool)
    convert_numbers_to_hangul_phonemes = _check_type(convert_numbers_to_hangul_phonemes,
                                                     "convert_numbers_to_hangul_phonemes", bool)
    verbose = _check_type(verbose, "verbose", bool)
    num_workers = _check_num_workers(text, num_workers)
    _check_backend_mecab_pecab_only(backend)

    return _run_job(
        func=partial(
            _g2p,
            descriptive=descriptive,
            group_vowels=group_vowels,
            to_syllable=to_syllable,
            convert_english_to_hangul_phonemes=convert_english_to_hangul_phonemes,
            convert_numbers_to_hangul_phonemes=convert_numbers_to_hangul_phonemes,
            backend=backend,
            verbose=verbose,
        ),
        inputs=text,
        num_workers=num_workers,
    )


def _g2p(
    text: str,
    descriptive: bool = False,
    group_vowels: bool = False,
    to_syllable: bool = True,
    convert_english_to_hangul_phonemes: bool = True,
    convert_numbers_to_hangul_phonemes: bool = True,
    backend: str = "auto",
    verbose: bool = False,
):
    """
    Main function

    Args:
        text: input text
        descriptive: boolean.
        group_vowels: boolean. If True, the vowels of the identical sound are normalized.
        to_syllable: boolean. If True, hangul letters or jamo are assembled to form syllables.
        convert_english_to_hangul_phonemes: boolean. If True, convert English to Hangul phonemes.
        convert_numbers_to_hangul_phonemes: boolean. If True, convert numbers to Hangul phonemes.
        backend: str
        verbose: boolean

    Notes:
        For example, given an input string "나의 친구가 mp3 file 3개를 다운받고 있다",
        STEP 1. idioms
        -> 나의 친구가 엠피쓰리 file 3개를 다운받고 있다

        STEP 2. English to Hangul
        -> 나의 친구가 엠피쓰리 파일 3개를 다운받고 있다

        STEP 3. annotate
        -> 나의/J 친구가 엠피쓰리 파일 3개/B를 다운받고 있다

        STEP 4. Spell out arabic numbers
        -> 나의/J 친구가 엠피쓰리 파일 세개/B를 다운받고 있다

        STEP 5. decompose
        -> 나의/J 친구가 엠피쓰리 파일 세개/B를 다운받고 있다

        STEP 6-9. Hangul
        -> 나의 친구가 엠피쓰리 파일 세개를 다운받꼬 읻따

    Returns:
        str: Phoneme string
    """
    # 1. idioms
    text = convert_idioms(
        text=text,
        descriptive=descriptive,
        convert_english_to_hangul_phonemes=convert_english_to_hangul_phonemes,
        convert_numbers_to_hangul_phonemes=convert_numbers_to_hangul_phonemes,
        verbose=verbose,
    )

    # 2 English to Hangul
    if convert_english_to_hangul_phonemes:
        text = convert_eng(text)

    # 3. annotate
    text = annotate(text, backend=backend)

    # 4. Spell out arabic numbers
    if convert_numbers_to_hangul_phonemes:
        text = convert_num(text)

    # 5. decompose
    inp = h2j(text)

    # 6. special
    for func in (jyeo, ye, consonant_ui, josa_ui, vowel_ui, \
                 jamo, rieulgiyeok, rieulbieub, verb_nieun, \
                 balb, palatalize, modifying_rieul):
        inp = func(inp, descriptive, verbose)
    inp = re.sub("/[PJEB]", "", inp)

    # 7. regular table: batchim + onset
    for str1, str2, rule_ids in table:
        _inp = inp
        inp = re.sub(str1, str2, inp)

        if len(rule_ids) > 0:
            rule = "\n".join(rule_id2text.get(rule_id, "") for rule_id in rule_ids)
        else:
            rule = ""
        gloss(verbose, inp, _inp, rule)

    # 8 link
    for func in (link1, link2, link3, link4):
        inp = func(inp, descriptive, verbose)

    # 9. postprocessing
    if group_vowels:
        inp = group(inp)

    if to_syllable:
        inp = j2h(inp, add_placeholder_for_leading_vowels=True)
    return inp



from kss import Kss


def test_qwerty():
    qwerty = Kss("qwerty")
    text = "dkssudgktpdy"
    output = qwerty(text, src="en", tgt="ko")
    assert output == '안녕하세요'
    text = "안녕하세요"
    output = qwerty(text, src="ko", tgt="en")
    assert output == 'dkssudgktpdy'



from kss import Kss


def test_romanize():
    romanize = Kss("romanize")
    text = "안녕하세요"
    output = romanize(text)
    assert output == 'annyeonghaseyo'
    text = "대관령"
    output = romanize(text)
    assert output == 'daegwallyeong'



# This code was copied from KR-WordRank [https://github.com/lovit/KR-WordRank]
# And modified by Hyunwoong Ko [https://github.com/hyuwoongko]

# Almost all code for wordrank is same with the original code.
# But there are unnecessary dependencies (scikit-learn, numpy, scipy).
# So I removed such dependencies for this project.

from collections import defaultdict

from kss._modules.morphemes.split_morphemes import split_morphemes


def hits(graph, beta, max_iter=50, bias=None, verbose=True,
         sum_weight=100, number_of_nodes=None, converge=0.001):
    """
    It trains rank of node using HITS algorithm.

    Arguments
    ---------
    graph : dict of dict
        Adjacent subword graph. graph[int][int] = float
    beta : float
        PageRank damping factor
    max_iter : int
        Maximum number of iterations
    bias : None or dict
        Bias vector
    verbose : Boolean
        If True, it shows training progress.
    sum_weight : float
        Sum of weights of all nodes in graph
    number_of_nodes : None or int
        Number of nodes in graph
    converge : float
        Minimum rank difference between previous step and current step.
        If the difference is smaller than converge, it do early-stop.

    Returns
    -------
    rank : dict
        Rank dictionary formed as {int:float}.
    """

    if not bias:
        bias = {}
    if not number_of_nodes:
        number_of_nodes = max(len(graph), len(bias))

    if number_of_nodes <= 1:
        raise ValueError(
            'The graph should consist of at least two nodes\n',
            'The node size of inserted graph is %d' % number_of_nodes
        )

    dw = sum_weight / number_of_nodes
    rank = {node: dw for node in graph.keys()}

    for num_iter in range(1, max_iter + 1):
        rank_ = _update(rank, graph, bias, dw, beta)
        diff = sum((abs(w - rank.get(n, 0)) for n, w in rank_.items()))
        rank = rank_

        if diff < sum_weight * converge:
            if verbose:
                print('\riter = %d Early stopped.' % num_iter, end='', flush=True)
            break

        if verbose:
            print('\riter = %d' % num_iter, end='', flush=True)

    if verbose:
        print('\rdone')

    return rank


def _update(rank, graph, bias, dw, beta):
    rank_new = {}
    for to_node, from_dict in graph.items():
        rank_new[to_node] = sum([w * rank[from_node] for from_node, w in from_dict.items()])
        rank_new[to_node] = beta * rank_new[to_node] + (1 - beta) * bias.get(to_node, dw)
    return rank_new


def summarize_with_keywords(texts, num_keywords=100, stopwords=None, min_count=5,
                            max_length=10, beta=0.85, max_iter=10, num_rset=-1, verbose=False):
    """
    It train KR-WordRank to extract keywords from texts.

        >>> from krwordrank.word import summarize_with_keywords

        >>> texts = [] # list of str
        >>> keywords = summarize_with_keywords(texts, num_keywords=100, min_count=5)

    Arguments
    ---------
    texts : list of str
        Each str is a sentence.
    num_keywords : int
        Number of keywords extracted from KR-WordRank
        Default is 100.
    stopwords : None or set of str
        Stopwords list for keyword and key-sentence extraction
    min_count : int
        Minimum frequency of subwords used to construct subword graph
        Default is 5
    max_length : int
        Maximum length of subwords used to construct subword graph
        Default is 10
    beta : float
        PageRank damping factor. 0 < beta < 1
        Default is 0.85
    max_iter : int
        Maximum number of iterations of HITS algorithm.
        Default is 10
    num_rset : int
        Number of R set words sorted by rank. It will be used to L-part word filtering.
        Default is -1.
    verbose : Boolean
        If True, it shows training status
        Default is False

    Returns
    -------
    keywords : dict
        Word : rank dictionary. keywords[str] = float

    Usage
    -----
        >>> from krwordrank.word import summarize_with_keywords

        >>> texts = [] # list of str
        >>> keywords = summarize_with_keywords(texts, num_keywords=100, min_count=5)
    """
    # train KR-WordRank
    wordrank_extractor = KRWordRank(
        min_count=min_count,
        max_length=max_length,
        verbose=verbose
    )

    keywords, rank, graph = wordrank_extractor.extract(texts,
                                                       beta, max_iter, num_rset=num_rset)

    # stopword filtering
    if stopwords is None:
        stopwords = {}
    keywords = {word: r for word, r in keywords.items() if not (word in stopwords)}

    # top rank filtering
    if num_keywords > 0:
        keywords = {word: r for word, r in sorted(keywords.items(), key=lambda x: -x[1])[:num_keywords]}

    return keywords


class KRWordRank:
    """Unsupervised Korean Keyword Extractor

    Implementation of Kim, H. J., Cho, S., & Kang, P. (2014). KR-WordRank:
    An Unsupervised Korean Word Extraction Method Based on WordRank.
    Journal of Korean Institute of Industrial Engineers, 40(1), 18-33.

    Arguments
    ---------
    min_count : int
        Minimum frequency of subwords used to construct subword graph
        Default is 5
    max_length : int
        Maximum length of subwords used to construct subword graph
        Default is 10
    verbose : Boolean
        If True, it shows training status
        Default is False

    Examples
    -----
        >>> from krwordrank.word import KRWordRank

        >>> texts = ['예시 문장 입니다', '여러 문장의 list of str 입니다', ... ]
        >>> wordrank_extractor = KRWordRank()
        >>> keywords, rank, graph = wordrank_extractor.extract(texts, beta, max_iter, verbose)
    """

    def __init__(self, min_count=5, max_length=10, backend="auto", noun_only=True, verbose=False):
        self.min_count = min_count
        self.max_length = max_length
        self.verbose = verbose
        self.backend = backend
        self.noun_only = noun_only
        self.sum_weight = 1
        self.vocabulary = {}
        self.index2vocab = []

    def scan_vocabs(self, docs):
        """
        It scans subwords positioned of left-side (L) and right-side (R) of words.
        After scanning was done, KR-WordRank has index2vocab as class attribute.

        Arguments
        ---------
        docs : list of str
            Sentence list

        Returns
        -------
        counter : dict
            {(subword, 'L')] : frequency}
        """
        self.vocabulary = {}
        if self.verbose:
            print('scan vocabs ... ')

        counter = {}
        for doc in docs:

            for (token, pos) in split_morphemes(doc, backend=self.backend):
                if self.noun_only:
                    if not pos.startswith('N'):
                        continue
                len_token = len(token)
                counter[(token, 'L')] = counter.get((token, 'L'), 0) + 1

                for e in range(1, min(len(token), self.max_length)):
                    if (len_token - e) > self.max_length:
                        continue

                    l_sub = (token[:e], 'L')
                    r_sub = (token[e:], 'R')
                    counter[l_sub] = counter.get(l_sub, 0) + 1
                    counter[r_sub] = counter.get(r_sub, 0) + 1

        counter = {token: freq for token, freq in counter.items() if freq >= self.min_count}
        for token, _ in sorted(counter.items(), key=lambda x: x[1], reverse=True):
            self.vocabulary[token] = len(self.vocabulary)

        self._build_index2vocab()

        if self.verbose:
            print('num vocabs = %d' % len(counter))
        return counter

    def _build_index2vocab(self):
        self.index2vocab = [vocab for vocab, index in sorted(self.vocabulary.items(), key=lambda x: x[1])]
        self.sum_weight = len(self.index2vocab)

    def extract(self, docs, beta=0.85, max_iter=10, num_keywords=-1,
                num_rset=-1, vocabulary=None, bias=None, rset=None):
        """
        It constructs word graph and trains ranks of each node using HITS algorithm.
        After training it selects suitable subwords as words.

        Arguments
        ---------
        docs : list of str
            Sentence list.
        beta : float
            PageRank damping factor. 0 < beta < 1
            Default is 0.85
        max_iter : int
            Maximum number of iterations of HITS algorithm.
            Default is 10
        num_keywords : int
            Number of keywords sorted by rank.
            Default is -1. If the vaule is negative, it returns all extracted words.
        num_rset : int
            Number of R set words sorted by rank. It will be used to L-part word filtering.
            Default is -1.
        vocabulary : None or dict
            User specified vocabulary to index mapper
        bias : None or dict
            User specified HITS bias term
        rset : None or dict
            User specfied R set

        Returns
        -------
        keywords : dict
            word : rank dictionary. {str:float}
        rank : dict
            subword : rank dictionary. {int:float}
        graph : dict of dict
            Adjacent subword graph. {int:{int:float}}

        Examples
        -----
            >>> from krwordrank.word import KRWordRank

            >>> texts = ['예시 문장 입니다', '여러 문장의 list of str 입니다', ... ]
            >>> wordrank_extractor = KRWordRank()
            >>> keywords, rank, graph = wordrank_extractor.extract(texts, beta, max_iter, verbose)
        """

        rank, graph = self.train(docs, beta, max_iter, vocabulary, bias)

        lset = {self.int2token(idx)[0]: r for idx, r in rank.items() if self.int2token(idx)[1] == 'L'}
        if not rset:
            rset = {self.int2token(idx)[0]: r for idx, r in rank.items() if self.int2token(idx)[1] == 'R'}

        if num_rset > 0:
            rset = {token: r for token, r in sorted(rset.items(), key=lambda x: -x[1])[:num_rset]}

        keywords = self._select_keywords(lset, rset)
        keywords = self._filter_compounds(keywords)
        keywords = self._filter_subtokens(keywords)

        if num_keywords > 0:
            keywords = {token: r for token, r in sorted(keywords.items(), key=lambda x: -x[1])[:num_keywords]}

        return keywords, rank, graph

    def train(self, docs, beta=0.85, max_iter=10, vocabulary=None, bias=None):
        """
        It constructs word graph and trains ranks of each node using HITS algorithm.
        Use this function only when you want to train rank of subwords

        Arguments
        ---------
        docs : list of str
            Sentence list.
        beta : float
            PageRank damping factor. 0 < beta < 1
            Default is 0.85
        max_iter : int
            Maximum number of iterations of HITS algorithm.
            Default is 10
        vocabulary : None or dict
            User specified vocabulary to index mapper
        bias : None or dict
            User specified HITS bias term
            {str: float} Format

        Returns
        -------
        rank : dict
            subword : rank dictionary. {int:float}
        graph : dict of dict
            Adjacent subword graph. {int:{int:float}}
        """
        if (not vocabulary) and (not self.vocabulary):
            self.scan_vocabs(docs)
        elif not vocabulary:
            self.vocabulary = vocabulary
            self._build_index2vocab()

        graph = self._construct_word_graph(docs)

        # add custom bias dict
        encoded_bias = {}
        custom_bias_dict = bias

        if custom_bias_dict:
            for word, value in custom_bias_dict.items():
                encoded_word = self.token2int((word, 'L'))
                if encoded_word != -1:
                    encoded_bias[encoded_word] = value

        rank = hits(graph, beta, max_iter, encoded_bias,
                    sum_weight=self.sum_weight,
                    number_of_nodes=len(self.vocabulary),
                    verbose=self.verbose
                    )

        return rank, graph

    def token2int(self, token):
        """
        Arguments
        ---------
        token : tuple
            (subword, 'L') or (subword, 'R')
            For example, ('이것', 'L') or ('은', 'R')

        Returns
        -------
        index : int
            Corresponding index
            If it is unknown, it returns -1
        """
        return self.vocabulary.get(token, -1)

    def int2token(self, index):
        """
        Arguments
        ---------
        index : int
            Token index

        Returns
        -------
        token : tuple
            Corresponding index formed such as (subword, 'L') or (subword, 'R')
            For example, ('이것', 'L') or ('은', 'R').
            If it is unknown, it returns None
        """
        return self.index2vocab[index] if (0 <= index < len(self.index2vocab)) else None

    def _construct_word_graph(self, docs):
        def normalize(graph):
            graph_ = defaultdict(lambda: defaultdict(lambda: 0))
            for from_, to_dict in graph.items():
                sum_ = sum(to_dict.values())
                for to_, w in to_dict.items():
                    graph_[to_][from_] = w / sum_
            graph_ = {t: dict(fd) for t, fd in graph_.items()}
            return graph_

        graph = defaultdict(lambda: defaultdict(lambda: 0))
        for doc in docs:

            tokens = doc.split()

            if not tokens:
                continue

            links = []
            for token in tokens:
                links += self._intra_link(token)

            if len(tokens) > 1:
                tokens = [tokens[-1]] + tokens + [tokens[0]]
                links += self._inter_link(tokens)

            links = self._check_token(links)
            if not links:
                continue

            links = self._encode_token(links)
            for l_node, r_node in links:
                graph[l_node][r_node] += 1
                graph[r_node][l_node] += 1

        # reverse for inbound graph. but it normalized with sum of outbound weight
        graph = normalize(graph)
        return graph

    def _check_token(self, token_list):
        return [(token[0], token[1]) for token in token_list if
                (token[0] in self.vocabulary and token[1] in self.vocabulary)]

    def _encode_token(self, token_list):
        return [(self.vocabulary[token[0]], self.vocabulary[token[1]]) for token in token_list]

    def _intra_link(self, token):
        links = []
        len_token = len(token)
        for e in range(1, min(len_token, 10)):
            if (len_token - e) > self.max_length:
                continue
            links.append(((token[:e], 'L'), (token[e:], 'R')))
        return links

    @staticmethod
    def _inter_link(tokens):
        def rsub_to_token(t_left, t_curr):
            return [((t_left[-b:], 'R'), (t_curr, 'L')) for b in range(1, min(10, len(t_left)))]

        def token_to_lsub(t_curr, t_rigt):
            return [((t_curr, 'L'), (t_rigt[:e], 'L')) for e in range(1, min(10, len(t_rigt)))]

        links = []
        for i in range(1, len(tokens) - 1):
            links += rsub_to_token(tokens[i - 1], tokens[i])
            links += token_to_lsub(tokens[i], tokens[i + 1])
        return links

    @staticmethod
    def _filter_subtokens(keywords):
        subtokens = set()
        keywords_ = {}

        for word, r in sorted(keywords.items(), key=lambda x: x[1], reverse=True):
            subs = {word[:e] for e in range(2, len(word) + 1)}

            is_subtoken = False
            for sub in subs:
                if sub in subtokens:
                    is_subtoken = True
                    break

            if not is_subtoken:
                keywords_[word] = r
                subtokens.update(subs)

        return keywords_

    @staticmethod
    def _select_keywords(lset, rset):
        keywords = {}
        for word, r in sorted(lset.items(), key=lambda x: x[1], reverse=True):
            len_word = len(word)
            if len_word == 1:
                continue

            is_compound = False
            for e in range(2, len_word):
                if (word[:e] in keywords) and (word[:e] in rset):
                    is_compound = True
                    break

            if not is_compound:
                keywords[word] = r

        return keywords

    @staticmethod
    def _filter_compounds(keywords):
        keywords_ = {}
        for word, r in sorted(keywords.items(), key=lambda x: x[1], reverse=True):
            len_word = len(word)

            if len_word <= 2:
                keywords_[word] = r
                continue

            if len_word == 3:
                if word[:2] in keywords_:
                    continue

            is_compound = False
            for e in range(2, len_word - 1):
                # fixed. comment from Y. cho
                if (word[:e] in keywords) and (word[e:] in keywords):
                    is_compound = True
                    break

            if not is_compound:
                keywords_[word] = r

        return keywords_



from typing import Union, List, Tuple

from kss._utils.multiprocessing import _run_job
from kss._utils.sanity_checks import _check_text, _check_num_workers

UNICODE_TO_REMOVE = {
    0x0000: None,  # Null
    0x0001: None,  # Start of Heading
    0x0002: None,  # Start of Text
    0x0003: None,  # End of Text
    0x0004: None,  # End of Transmission
    0x0005: None,  # Enquiry
    0x0006: None,  # Acknowledge
    0x0007: None,  # Bell
    0x0008: None,  # Backspace
    0x000B: None,  # Vertical Tab
    0x000C: None,  # Form Feed
    0x000D: None,  # Carriage Return
    0x000E: None,  # Shift Out
    0x000F: None,  # Shift In
    0x0010: None,  # Data Link Escape
    0x0011: None,  # Device Control 1
    0x0012: None,  # Device Control 2
    0x0013: None,  # Device Control 3
    0x0014: None,  # Device Control 4
    0x0015: None,  # Negative Acknowledge
    0x0016: None,  # Synchronous Idle
    0x0017: None,  # End of Transmission Block
    0x0018: None,  # Cancel
    0x0019: None,  # End of Medium
    0x001A: None,  # Substitute
    0x001B: None,  # Escape
    0x001C: None,  # File Separator
    0x001D: None,  # Group Separator
    0x001E: None,  # Record Separator
    0x001F: None,  # Unit Separator
    0x007F: None,  # Delete
    0x0080: None,  # Padding Character
    0x0081: None,  # High Octet Preset
    0x0082: None,  # Break Permitted Here
    0x0083: None,  # No Break Here
    0x0084: None,  # Index
    0x0085: None,  # Next Line
    0x0086: None,  # Start of Selected Area
    0x0087: None,  # End of Selected Area
    0x0088: None,  # Character Tabulation Set
    0x0089: None,  # Character Tabulation with Justification
    0x008A: None,  # Line Tabulation Set
    0x008B: None,  # Partial Line Forward
    0x008C: None,  # Partial Line Backward
    0x008D: None,  # Reverse Line Feed
    0x008E: None,  # Single-Shift Two
    0x008F: None,  # Single-Shift Three
    0x0090: None,  # Device Control String
    0x0091: None,  # Private Use 1
    0x0092: None,  # Private Use 2
    0x0093: None,  # Set Transmit State
    0x0094: None,  # Cancel Character
    0x0095: None,  # Message Waiting
    0x0096: None,  # Start of Guarded Area
    0x0097: None,  # End of Guarded Area
    0x0098: None,  # Start of String
    0x0099: None,  # Single Graphic Character Introducer
    0x009A: None,  # Single Character Introducer
    0x009B: None,  # Control Sequence Introducer
    0x009C: None,  # String Terminator
    0x009D: None,  # Operating System Command
    0x009E: None,  # Privacy Message
    0x009F: None,  # Application Program Command
    0x00A0: None,  # No-Break Space
    0x00AD: None,  # Soft Hyphen
    0x115f: None,  # Hangul Choseong Filler
    0x1160: None,  # Hangul Jungseong Filler
    0x200B: None,  # Zero Width Space
    0x200E: None,  # Left-to-Right Mark
    0x200F: None,  # Right-to-Left Mark
    0x202A: None,  # Left-to-Right Embedding
    0x202B: None,  # Right-to-Left Embedding
    0x202C: None,  # Pop Directional Formatting
    0x202D: None,  # Left-to-Right Override
    0x202E: None,  # Right-to-Left Override
    0x202F: None,  # Narrow No-Break Space
    0x2061: None,  # Function Application
    0x2062: None,  # Invisible Times
    0x2063: None,  # Invisible Separator
    0x2064: None,  # Invisible Plus
    0x2066: None,  # Left-to-Right Isolate
    0x2067: None,  # Right-to-Left Isolate
    0x2068: None,  # First Strong Isolate
    0x2069: None,  # Pop Directional Isolate
    0x206A: None,  # Inhibit Symmetric Swapping
    0x206B: None,  # Activate Symmetric Swapping
    0x206E: None,  # National Digit Shapes
    0x206F: None,  # Nominal Digit Shapes
    0x3164: None,  # Hangul Filler
    0xFEFF: None,  # Zero Width No-Break Space
    0xFFA0: None,  # Halfwidth Hangul Filler
    0xFFFC: None,  # Object Replacement Character
    0xFFFE: None,  # Byte Order Mark
    0xFFFF: None,  # Non character
    0x1BCA0: None,  # Shorthand Format Letter Overlap
}
UNICODE_TO_REMOVE.update({char: None for char in range(0xE0000, 0xF8FF + 1)})  # Tag + PUA characters
UNICODE_TO_REMOVE.update({char: None for char in range(0xF0000, 0xFFFFF + 1)})  # PUA characters
UNICODE_TO_REMOVE.update({char: None for char in range(0x100000, 0x10FFFF + 1)})  # PUA characters
UNICODE_TO_REMOVE.update({char: None for char in range(0x1D100, 0x1D1FF + 1)})  # Musical symbols

UNICODE_TO_REPLACE = {
    0x2028: "\n",  # Line Separator
    0x2029: "\n\n",  # Paragraph Separator
    0x2000: " ",  # En Quad
    0x2001: " ",  # Em Quad
    0x2002: " ",  # En Space
    0x2003: " ",  # Em Space
    0x2004: " ",  # Three-Per-Em Space
    0x2005: " ",  # Four-Per-Em Space
    0x2006: " ",  # Six-Per-Em Space
    0x2007: " ",  # Figure Space
    0x2008: " ",  # Punctuation Space
    0x2009: " ",  # Thin Space
    0x200A: " ",  # Hair Space
    0x205F: " ",  # Medium Mathematical Space
    0x3000: " ",  # Ideographic Space
}


# The following characters are also invisible, but they should not be removed or replaced.
# Because they affect the meaning of the text. In this case, we don't need to do anything,
# but I just wrote this for documentation.

# UNICODE_TO_REMAIN = {
#     0x0020: " ",  # Space
#     0x0009: "\t",  # Horizontal Tab
#     0x000A: "\n",  # Line Feed
#     0x034F: "\u034F",  # Combining Grapheme Joiner
#     0x061C: "\u061C",  # Arabic Letter Mark
#     0x1680: "\u1680",  # Ogham Space Mark
#     0x17B4: "\u17B4",  # Khmer Vowel Inherent AQ
#     0x17B5: "\u17B5",  # Khmer Vowel Inherent AA
#     0x180B: "\u180B",  # Mongolian Free Variation Selector One
#     0x180C: "\u180C",  # Mongolian Free Variation Selector Two
#     0x180D: "\u180D",  # Mongolian Free Variation Selector Three
#     0x180E: "\u180E",  # Mongolian Vowel Separator
#     0x2800: "\u2800",  # Braille Pattern Blank
#     0x200C: "\u200C",  # Zero Width Non-Joiner
#     0x200D: "\u200D",  # Zero Width Joiner
#     0x2060: "\u2060",  # Word Joiner
#     0x206C: "\u206C,   # Inhibit Arabic Form Shaping
#     0x206D: "\u206D",  # Activate Arabic Form Shaping
#     0x1307B: "\U0001307B",  # Egyptian Hieroglyph Alef
# }
# UNICODE_TO_REMAIN.update({char: chr(char) for char in range(0x02B0, 0x02FF + 1)})  # Modifier letters
# UNICODE_TO_REMAIN.update({char: chr(char) for char in range(0x0300, 0x036F + 1)})  # Combining Diacritical Marks
# UNICODE_TO_REMAIN.update({char: chr(char) for char in range(0xFE00, 0xFE0F + 1)})  # Variation Selectors
# UNICODE_TO_REMAIN.update({char: chr(char) for char in range(0x1E0100, 0x1E01EF + 1)})  # Variation Selectors


def remove_invisible_chars(
    text: Union[str, List[str], Tuple[str]],
    num_workers: Union[int, str] = "auto",
) -> Union[str, List[str]]:
    """
    This removes invisible characters from text.

    Args:
        text (Union[str, List[str], Tuple[str]]): single text or list of texts
        num_workers (Union[int, str]): the number of multiprocessing workers

    Returns:
        Union[str, List[str]]: text with removed invisible characters or list of texts with removed invisible characters

    Examples:
        >>> from kss import Kss
        >>> remove_invisible_chars = Kss("remove_invisible_chars")
        >>> text = "안녕\u200b하세요"
        >>> remove_invisible_chars(text)
        '안녕하세요'
    """

    text, finish = _check_text(text)

    if finish:
        return text

    num_workers = _check_num_workers(text, num_workers)

    return _run_job(
        func=_remove_invisible_chars,
        inputs=text,
        num_workers=num_workers,
    )


def _remove_invisible_chars(text: str) -> str:
    text = text.translate(UNICODE_TO_REMOVE)
    text = text.translate(UNICODE_TO_REPLACE)
    return text



# This code was copied from KR-WordRank [https://github.com/lovit/KR-WordRank]
# And modified by Hyunwoong Ko [https://github.com/hyuwoongko]
from typing import List, Union, Tuple

from kss._modules.keywords.utils import KRWordRank
from kss._utils.sanity_checks import _check_text, _check_type, _check_backend_mecab_pecab_only


def extract_keywords(
    text: Union[str, List[str]],
    num_keywords: int = 5,
    min_word_count: int = 1,
    max_word_length: int = 10,
    return_scores: bool = False,
    backend: str = "auto",
    noun_only: bool = True,
    num_workers: Union[int, str] = "auto",
) -> Union[List[str], List[Tuple[str, float]]]:
    """
    This extracts keywords from the given text.
    This uses TextRank algorithm to extract keywords.

    Args:
        text (Union[str, List[str]): single text or list of texts
        num_keywords (int): the number of keywords to extract
        min_word_count (int): the minimum count of words
        max_word_length (int): the maximum length of words
        return_scores (bool): whether to return scores or not
        backend (str): morpheme analyzer backend. 'mecab', 'pecab' are supported
        noun_only (bool): whether to extract only nouns or not
        num_workers (Union[int, str]): the number of multiprocessing workers

    Returns:
        Union[List[str], List[Tuple[str, float]]]: list of keywords or list of tuples of keywords and scores

    Examples:
        >>> from kss import Kss
        >>> extract_keywords = Kss("extract_keywords")
        >>> text = ['여운이 크게남는영화 엠마스톤 너무 사랑스럽고 라이언고슬링 남자가봐도 정말 매력적인 배우인듯 영상미 음악 연기 구성 전부 좋았고 마지막 엔딩까지 신선하면서 애틋하구요 30중반에 감정이 많이 메말라있었는데 오랜만에 가슴이 촉촉해지네요',
        ...         '영상미도 너무 아름답고 신나는 음악도 좋았다 마지막 세바스찬과 미아의 눈빛교환은 정말 마음 아팠음 영화관에 고딩들이 엄청 많던데 고딩들은 영화 내용 이해를 못하더라ㅡㅡ사랑을 깊게 해본 사람이라면 누구나 느껴볼수있는 먹먹함이 있다',
        ...         '정말 영상미랑 음악은 최고였다 그리고 신선했다 음악이 너무 멋있어서 연기를 봐야 할지 노래를 들어야 할지 모를 정도로 그리고 보고 나서 생각 좀 많아진 영화 정말 이 연말에 보기 좋은 영화 인 것 같다',
        ...         '무언의 마지막 피아노연주 완전 슬픔ㅠ보는이들에게 꿈을 상기시켜줄듯 또 보고 싶은 내생에 최고의 뮤지컬영화였음 단순할수 있는 내용에 뮤지컬을 가미시켜째즈음악과 춤으로 지루할틈없이 빠져서봄 ost너무좋았음',
        ...         '처음엔 초딩들 보는 그냥 그런영화인줄 알았는데 정말로 눈과 귀가 즐거운 영화였습니다 어찌보면 뻔한 스토리일지 몰라도 그냥 보고 듣는게 즐거운 그러다가 정말 마지막엔 너무 아름답고 슬픈 음악이 되어버린',
        ...         '정말 멋진 노래와 음악과 영상미까지 정말 너무 멋있는 영화 눈물을 흘리면서 봤습니다 영화가 끝난 순간 감탄과 동시에 여운이 길게 남아 또 눈물을 흘렸던내 인생 최고의 뮤지컬 영화',
        ...         '평소 뮤지컬 영화 좋아하는 편인데도 평점에 비해 너무나 별로였던 영화 재즈음악이나 영상미 같은 건 좋았지만 줄거리도 글쎄 결말은 정말 별로 6 7점 정도 주는게 맞다고 생각하지만 개인적으로 후반부가 너무 별로여서',
        ...         '오랜만에 좋은 영화봤다는 생각들었구요 음악도 영상도 스토리도 너무나좋았고 무엇보다 진한 여운이 남는 영화는 정말 오랜만이었어요 연인끼리 가서 보기 정말 좋은영화 너뮤너뮤너뮤 재밌게 잘 봤습니다',
        ...         '음악 미술 연기 등 모든 것이 좋았지만 마지막 결말이 너무 현실에 뒤떨어진 꿈만 같다 꿈을 이야기하는 영화지만 과정과 결과에 있어 예술가들의 현실을 너무 반영하지 못한 것이 아닌가하는 생각이든다 그래서 보고 난 뒤 나는 꿈을 꿔야하는데 허탈했다',
        ...         '마지막 회상씬의 감동이 잊혀지질않는다마지막 십분만으로 티켓값이 아깝지않은 영화 음악들도 너무 아름다웠다옛날 뮤지컬 같은 빈티지영상미도 최고']
        >>> output = extract_keywords(text, noun_only=True)
        >>> print(output)
        ['마지막', '영화', '음악', '라이언고슬링', '뮤지컬']
        >>> output = extract_keywords(text, noun_only=False)
        >>> print(output)
        ['너무', '정말', '마지막', '영화', '음악']

    References:
        This was copied from [KR-WordRank](https://github.com/lovit/KR-WordRank) and modified by Kss
    """

    text, finish = _check_text(text)

    if finish:
        return []

    num_keywords = _check_type(num_keywords, "num_keywords", int)
    min_word_count = _check_type(min_word_count, "min_word_count", int)
    max_word_length = _check_type(max_word_length, "max_word_length", int)
    return_scores = _check_type(return_scores, "return_scores", bool)
    noun_only = _check_type(noun_only, "noun_only", bool)
    _check_backend_mecab_pecab_only(backend)

    if num_workers != "auto":
        raise ValueError("`extract_keywords` does not support `num_workers` argument")

    return _extract_keywords(
        text=text,
        num_keywords=num_keywords,
        min_word_count=min_word_count,
        max_word_length=max_word_length,
        return_scores=return_scores,
        noun_only=noun_only,
        backend=backend,
    )


def _extract_keywords(
    text: Union[str, List[str]],
    num_keywords: int = 5,
    min_word_count: int = 1,
    max_word_length: int = 10,
    return_scores: bool = False,
    backend: str = "auto",
    noun_only: bool = True,
):
    if isinstance(text, str):
        text = [text]

    wordrank_extractor = KRWordRank(
        min_count=min_word_count,
        max_length=max_word_length,
        backend=backend,
        noun_only=noun_only,
    )
    keywords = wordrank_extractor.extract(
        text, num_keywords=num_keywords
    )[0]

    if return_scores:
        return keywords
    else:
        return list(keywords.keys())



from typing import Union, List, Tuple

from kss._utils.multiprocessing import _run_job
from kss._utils.sanity_checks import _check_num_workers, _check_text


replace_map = {
    "ﾡ": "ㄱ",
    "ﾢ": "ㄲ",
    "ﾣ": "ㄳ",
    "ﾤ": "ㄴ",
    "ﾥ": "ㄵ",
    "ﾦ": "ㄶ",
    "ﾧ": "ㄷ",
    "ﾨ": "ㄸ",
    "ﾩ": "ㄹ",
    "ﾪ": "ㄺ",
    "ﾫ": "ㄻ",
    "ﾬ": "ㄼ",
    "ﾭ": "ㄽ",
    "ﾮ": "ㄾ",
    "ﾯ": "ㄿ",
    "ﾱ": "ㅁ",
    "ﾲ": "ㅂ",
    "ﾳ": "ㅃ",
    "ﾴ": "ㅄ",
    "ﾵ": "ㅅ",
    "ﾶ": "ㅆ",
    "･": "ㆍ",
    "ﾷ": "ㅇ",
    "ﾸ": "ㅈ",
    "ﾹ": "ㅉ",
    "ﾺ": "ㅊ",
    "ﾻ": "ㅋ",
    "ﾼ": "ㅌ",
    "ﾽ": "ㅍ",
    "ﾾ": "ㅎ",
    "ￂ": "ㅏ",
    "ￃ": "ㅐ",
    "ￄ": "ㅑ",
    "ￆ": "ㅓ",
    "ￇ": "ㅔ",
    "ￊ": "ㅕ",
    "ￌ": "ㅗ",
    "ￏ": "ㅚ",
    "ￒ": "ㅛ",
    "ￓ": "ㅜ",
    "ￖ": "ㅟ",
    "ￗ": "ㅠ",
    "ￚ": "ㅡ",
    "ￜ": "ㅣ",
    "ￍ": "ㅘ",
    "ￔ": "ㅝ",
    "ￎ": "ㅙ",
    "ￕ": "ㅞ",
}


def half2full(
    text: Union[str, List[str], Tuple[str]],
    num_workers: Union[int, str] = "auto",
) -> Union[str, List[str]]:
    """
    This converts half-width characters to full-width characters.

    Args:
        text (Union[str, List[str], Tuple[str]]): single text or list of texts
        num_workers (Union[int, str]): the number of multiprocessing workers

    Returns:
        Union[str, List[str]]: converted text or list of converted texts

    Examples:
        >>> from kss import Kss
        >>> half2full = Kss("half2full")
        >>> text = "ﾻﾻﾻﾻﾻﾻ"
        >>> half2full(text)
        'ㅋㅋㅋㅋㅋㅋ'
    """
    text, finish = _check_text(text)

    if finish:
        return text

    num_workers = _check_num_workers(text, num_workers)

    return _run_job(
        func=_half2full,
        inputs=text,
        num_workers=num_workers,
    )


def _half2full(text: str):
    fullwidth_text = ""

    for char in text:
        fullwidth_text += replace_map.get(char, char)

    return fullwidth_text



# This is copied from dps [https://github.com/EleutherAI/dps]
# And modified by Hyunwoong Ko [https://github.com/hyunwoongko]

import re
from functools import partial
from typing import Tuple, List, Union

from kss._utils.multiprocessing import _run_job
from kss._utils.sanity_checks import _check_type, _check_num_workers, _check_text


PHONE_NUMBER_PATTERN = re.compile(r"([0-9]{2,3}-[0-9]{3,4}-[0-9]{4})|([0-9]{2,3}[0-9]{3,4}[0-9]{4})")

RRN_PATTERN = re.compile(r"([0-9]{6}-[0-9]{7})")

CARD_PATTERN = re.compile(r"([0-9]{4}-[0-9]{4}-[0-9]{4}-[0-9]{4})")

EMAIL_PATTERN = re.compile(
    r"[a-z0-9.\-+_]+@[a-z0-9.\-+_]+\.[a-z]+|[a-z0-9.\-+_]+@[a-z0-9.\-+_]+\.[a-z]+\.[a-z]"
)

BANK_ACCOUNT_PATTERN = re.compile(
    r"([0-9]\d{13})|([0-9,\-]{3,6}\-[0-9,\-]{2,6}\-[0-9,\-]{2,6})"  # IBK
    + r"([0-9]\d{13})|([0-9,\-]{3,6}\-[0-9,\-]{2,6}\-[0-9,\-]{2,6})"  # KB
    + r"([0-9]\d{12})|([0-9,\-]{3,6}\-[0-9,\-]{2,6}\-[0-9,\-]{2,6}\-[0-9,\-]{2})"  # NH
    + r"([0-9]\d{11})|([0-9,\-]{3,6}\-[0-9,\-]{2,6}\-[0-9,\-]{2,6})"  # SHINHAN
    + r"([0-9]\d{12})|([0-9,\-]{3,6}\-[0-9,\-]{2,6}\-[0-9,\-]{2,6})"  # WOORI
    + r"([0-9]\d{13})|([0-9,\-]{3,6}\-[0-9,\-]{2,6}\-[0-9,\-]{2,6})"  # KEB
    + r"([0-9]\d{11})|([0-9,\-]{3,6}\-[0-9,\-]{2,6}\-[0-9,\-]{2,6})"  # CITI
    + r"([0-9]\d{11})|([0-9]\d{11})|([0-9,\-]{3,6}\-[0-9,\-]{2,6}\-[0-9,\-]{2,6}\-[0-9,\-])"  # DGB
    + r"([0-9]\d{12})|([0-9]\d{12})|([0-9,\-]{3,6}\-[0-9,\-]{2,6}\-[0-9,\-]{2,6}\-[0-9,\-]{2})"  # BNK
    + r"([0-9]\d{10})|([0-9]\d{10})|([0-9,\-]{3,6}\-[0-9,\-]{2,6}\-[0-9,\-]{2,6})"  # SC
    + r"([0-9]\d{11})|([0-9]\d{10})|([0-9,\-]{3,6}\-[0-9,\-]{2,6}\-[0-9,\-]{2,6})"  # KBANK
    + r"([0-9]\d{12})|([0-9]\d{10})|([0-9,\-]{3,6}\-[0-9,\-]{2,6}\-[0-9,\-]{2,7})"  # KAKAO
)

CREDIT_CARD_PATTERN = re.compile(
    r"^(4026|417500|4405|4508|4844|4913|4917)\d+$|"  # VISA Electron
    + r"^(?:50|5[6-9]|6[0-9])\d+$|"  # Maestro
    + r"^(5019|4571)\d+$|"  # Dankort
    + r"^(62|81)\d+$|"  # China UnionPay
    + r"^4[0-9]\d+$|"  # Visa
    + r"^(?:5[1-5]|222[1-9]|22[3-9][0-9]|2[3-6][0-9][0-9]|27[0-1][0-9]|2720)\d+$|"  # MasterCard
    + r"^(34|37)\d+$|"  # American Express
    + r"^6(?:011|22(12[6-9]|1[3-9][0-9]|[2-8][0-9][0-9]|9[01][0-9]|92[0-5])|5|4|2[4-6][0-9]{3}|28[2-8][0-9]{2})\d+$|"  # Discover
    + r"^(35[2-8][0-9])\d+$|"  # JCB
    + r"^(636)\d+$|"  # InterPayment
    + r"^9[0-9]\d+$|"  # KOREAN
    + r"^(220[0-4])\d+$"  # MIR
)

ZIP_PATTERN = re.compile(r"\d{3}-\d{3}")

_base58_alphabet = '123456789ABCDEFGHJKLMNPQRSTUVWXYZabcdefghijkmnopqrstuvwxyz'
_bech32_alphabet = 'qpzry9x8gf2tvdw0s3jn54khce6mua7l'
BITCOIN_PATTERN = (
    r"( [13] ["
    + _base58_alphabet
    + "]{25,34}"
    + "| bc1 ["
    + _bech32_alphabet
    + "]{8,87})"
)

IPV4_PATTERN = r'(?:0|25[0-5]|2[0-4]\d|1\d?\d?|[1-9]\d?)(?:\.(?:0|25[0-5]|2[0-4]\d|1\d?\d?|[1-9]\d?)){3}'
IPV6_PATTERN = r'\[?((([0-9A-Fa-f]{1,4}:){7}([0-9A-Fa-f]{1,4}|:))|(([0-9A-Fa-f]{1,4}:){6}(:[0-9A-Fa-f]{1,' \
               r'4}|((25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)(\.(25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)){3})|:))|(([0-9A-Fa-f]{' \
               r'1,4}:){5}(((:[0-9A-Fa-f]{1,4}){1,2})|:((25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)(\.(25[0-5]|2[' \
               r'0-4]\d|1\d\d|[1-9]?\d)){3})|:))|(([0-9A-Fa-f]{1,4}:){4}(((:[0-9A-Fa-f]{1,4}){1,' \
               r'3})|((:[0-9A-Fa-f]{1,4})?:((25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)(\.(25[0-5]|2[0-4]\d|1\d\d|[' \
               r'1-9]?\d)){3}))|:))|(([0-9A-Fa-f]{1,4}:){3}(((:[0-9A-Fa-f]{1,4}){1,4})|((:[0-9A-Fa-f]{1,4}){0,' \
               r'2}:((25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)(\.(25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)){3}))|:))|(([' \
               r'0-9A-Fa-f]{1,4}:){2}(((:[0-9A-Fa-f]{1,4}){1,5})|((:[0-9A-Fa-f]{1,4}){0,3}:((25[0-5]|2[' \
               r'0-4]\d|1\d\d|[1-9]?\d)(\.(25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)){3}))|:))|(([0-9A-Fa-f]{1,4}:){1}(((:[' \
               r'0-9A-Fa-f]{1,4}){1,6})|((:[0-9A-Fa-f]{1,4}){0,4}:((25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)(\.(25[0-5]|2[' \
               r'0-4]\d|1\d\d|[1-9]?\d)){3}))|:))|(:(((:[0-9A-Fa-f]{1,4}){1,7})|((:[0-9A-Fa-f]{1,4}){0,' \
               r'5}:((25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)(\.(25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)){3}))|:)))(%.+)?\]?'

ul = '\u00a1-\uffff'  # Unicode letters range
_HOSTNAME_PATTERN = r'[a-z' + ul + r'0-9](?:[a-z' + ul + r'0-9-]{0,61}[a-z' + ul + r'0-9])?'
# Max length for domain name labels is 63 characters per RFC 1034 sec. 3.1
_DOMAIN_PATTERN = r'(?:\.(?!-)[a-z' + ul + r'0-9-]{1,63}(?<!-))*'
_TLD_PATTERN = (
    r'\.'  # dot
    r'(?!-)'  # can't start with a dash
    r'(?:[a-z' + ul + '-]{2,63}'  # domain label
                      r'|xn--[a-z0-9]{1,59})'  # or punycode label
                      r'(?<!-)'  # can't end with a dash
                      r'\.?'  # may have a trailing dot
)

_HOST_PATTERN = '(' + _HOSTNAME_PATTERN + _DOMAIN_PATTERN + _TLD_PATTERN + '|localhost)'

URL_PATTERN_1 = re.compile(
    r"""\b((?:https?://)?(?:(?:www\.)?(?:[\da-z\.-]+)\.(?:[a-z]{2,6})|(?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)|(?:(?:[0-9a-fA-F]{1,4}:){7,7}[0-9a-fA-F]{1,4}|(?:[0-9a-fA-F]{1,4}:){1,7}:|(?:[0-9a-fA-F]{1,4}:){1,6}:[0-9a-fA-F]{1,4}|(?:[0-9a-fA-F]{1,4}:){1,5}(?::[0-9a-fA-F]{1,4}){1,2}|(?:[0-9a-fA-F]{1,4}:){1,4}(?::[0-9a-fA-F]{1,4}){1,3}|(?:[0-9a-fA-F]{1,4}:){1,3}(?::[0-9a-fA-F]{1,4}){1,4}|(?:[0-9a-fA-F]{1,4}:){1,2}(?::[0-9a-fA-F]{1,4}){1,5}|[0-9a-fA-F]{1,4}:(?:(?::[0-9a-fA-F]{1,4}){1,6})|:(?:(?::[0-9a-fA-F]{1,4}){1,7}|:)|fe80:(?::[0-9a-fA-F]{0,4}){0,4}%[0-9a-zA-Z]{1,}|::(?:ffff(?::0{1,4}){0,1}:){0,1}(?:(?:25[0-5]|(?:2[0-4]|1{0,1}[0-9]){0,1}[0-9])\.){3,3}(?:25[0-5]|(?:2[0-4]|1{0,1}[0-9]){0,1}[0-9])|(?:[0-9a-fA-F]{1,4}:){1,4}:(?:(?:25[0-5]|(?:2[0-4]|1{0,1}[0-9]){0,1}[0-9])\.){3,3}(?:25[0-5]|(?:2[0-4]|1{0,1}[0-9]){0,1}[0-9])))(?::[0-9]{1,4}|[1-5][0-9]{4}|6[0-4][0-9]{3}|65[0-4][0-9]{2}|655[0-2][0-9]|6553[0-5])?(?:/[\w\.-]*)*/?)\b"""
)
URL_PATTERN_2 = re.compile(
    r'([a-z0-9.+-]*:?//)?'  # scheme is validated separately
    r'(?:[^\s:@/]+(?::[^\s:@/]*)?@)?'  # user:pass authentication
    r'(?:' + IPV4_PATTERN + '|' + IPV6_PATTERN + '|' + _HOST_PATTERN + ')'
                                                                       r'(?::\d{2,5})?'  # port
                                                                       r'(?:[/?#][^\s]*)?',  # resource path
    re.IGNORECASE
)


def anonymize(
    text: Union[str, List[str], Tuple[str]],
    phone_number_anonymization: bool = True,
    rrn_anonymization: bool = True,
    card_anonymization: bool = True,
    email_anonymization: bool = True,
    bank_account_anonymization: bool = True,
    credit_card_anonymization: bool = True,
    zip_anonymization: bool = True,
    bitcoin_anonymization: bool = True,
    url_anonymization: bool = True,
    ip_v6_anonymization: bool = True,
    ip_v4_anonymization: bool = True,
    phone_number_replacement: str = "<PHONE_NUMBER>",
    rrn_replacement: str = "<RRN>",
    card_replacement: str = "<CARD>",
    email_replacement: str = "<EMAIL>",
    bank_account_replacement: str = "<BANK_ACCOUNT>",
    credit_card_replacement: str = "<CREDIT_CARD>",
    zip_replacement: str = "<ZIP>",
    bitcoin_replacement: str = "<BITCOIN>",
    url_replacement: str = "<URL>",
    ip_v6_replacement: str = "<IPV6>",
    ip_v4_replacement: str = "<IPV4>",
    num_workers: Union[int, str] = "auto",
):
    """
    This anonymizes sensitive information in the given text.

    Args:
        text (Union[str, List[str], Tuple[str]): single text or list of texts
        phone_number_anonymization (bool): whether to anonymize phone numbers or not
        rrn_anonymization (bool): whether to anonymize resident registration numbers or not
        card_anonymization (bool): whether to anonymize card numbers or not
        email_anonymization (bool): whether to anonymize email addresses or not
        bank_account_anonymization (bool): whether to anonymize bank account numbers or not
        credit_card_anonymization (bool): whether to anonymize credit card numbers or not
        zip_anonymization (bool): whether to anonymize zip codes or not
        bitcoin_anonymization (bool): whether to anonymize bitcoin addresses or not
        url_anonymization (bool): whether to anonymize URLs or not
        ip_v6_anonymization (bool): whether to anonymize IPv6 addresses or not
        ip_v4_anonymization (bool): whether to anonymize IPv4 addresses or not
        phone_number_replacement (str): the replacement string for phone numbers, default is "<PHONE_NUMBER>"
        rrn_replacement (str): the replacement string for resident registration numbers, default is "<RRN>"
        card_replacement (str): the replacement string for card numbers, default is "<CARD>"
        email_replacement (str): the replacement string for email addresses, default is "<EMAIL>"
        bank_account_replacement (str): the replacement string for bank account numbers, default is "<BANK_ACCOUNT>"
        credit_card_replacement (str): the replacement string for credit card numbers, default is "<CREDIT_CARD>"
        zip_replacement (str): the replacement string for zip codes, default is "<ZIP>"
        bitcoin_replacement (str): the replacement string for bitcoin addresses, default is "<BITCOIN>"
        url_replacement (str): the replacement string for URLs, default is "<URL>"
        ip_v6_replacement (str): the replacement string for IPv6 addresses, default is "<IPV6>"
        ip_v4_replacement (str): the replacement string for IPv4 addresses, default is "<IPV4>"
        num_workers (Union[int, str]): the number of multiprocessing workers

    Returns:
        Union[str, List[str], Tuple[str]]: anonymized text or list of anonymized texts

    Examples:
        >>> from kss import Kss
        >>> anonymize = Kss("anonymize")
        >>> text = "제 전화번호는 010-1234-5678, 이메일 주소는 kevin.brain@kakaobrain.com입니다."
        >>> output = anonymize(text)
        >>> print(output)
        "제 전화번호는 <PHONE_NUMBER>, 이메일 주소는 <EMAIL>입니다."
    """
    text, finish = _check_text(text)

    if finish:
        return text

    phone_number_anonymization = _check_type(phone_number_anonymization, "phone_number_anonymization", bool)
    rrn_anonymization = _check_type(rrn_anonymization, "rrn_anonymization", bool)
    card_anonymization = _check_type(card_anonymization, "card_anonymization", bool)
    email_anonymization = _check_type(email_anonymization, "email_anonymization", bool)
    bank_account_anonymization = _check_type(bank_account_anonymization, "bank_account_anonymization", bool)
    credit_card_anonymization = _check_type(credit_card_anonymization, "credit_card_anonymization", bool)
    zip_anonymization = _check_type(zip_anonymization, "zip_anonymization", bool)
    bitcoin_anonymization = _check_type(bitcoin_anonymization, "bitcoin_anonymization", bool)
    url_anonymization = _check_type(url_anonymization, "url_anonymization", bool)
    ip_v6_anonymization = _check_type(ip_v6_anonymization, "ip_v6_anonymization", bool)
    ip_v4_anonymization = _check_type(ip_v4_anonymization, "ip_v4_anonymization", bool)
    phone_number_replacement = _check_type(phone_number_replacement, "phone_number_replacement", str)
    rrn_replacement = _check_type(rrn_replacement, "rrn_replacement", str)
    card_replacement = _check_type(card_replacement, "card_replacement", str)
    email_replacement = _check_type(email_replacement, "email_replacement", str)
    bank_account_replacement = _check_type(bank_account_replacement, "bank_account_replacement", str)
    credit_card_replacement = _check_type(credit_card_replacement, "credit_card_replacement", str)
    zip_replacement = _check_type(zip_replacement, "zip_replacement", str)
    bitcoin_replacement = _check_type(bitcoin_replacement, "bitcoin_replacement", str)
    url_replacement = _check_type(url_replacement, "url_replacement", str)
    ip_v6_replacement = _check_type(ip_v6_replacement, "ip_v6_replacement", str)
    ip_v4_replacement = _check_type(ip_v4_replacement, "ip_v4_replacement", str)
    num_workers = _check_num_workers(text, num_workers)

    return _run_job(
        func=partial(
            _anonymize,
            phone_number_anonymization=phone_number_anonymization,
            rrn_anonymization=rrn_anonymization,
            card_anonymization=card_anonymization,
            email_anonymization=email_anonymization,
            bank_account_anonymization=bank_account_anonymization,
            credit_card_anonymization=credit_card_anonymization,
            zip_anonymization=zip_anonymization,
            bitcoin_anonymization=bitcoin_anonymization,
            url_anonymization=url_anonymization,
            ip_v6_anonymization=ip_v6_anonymization,
            ip_v4_anonymization=ip_v4_anonymization,
            phone_number_replacement=phone_number_replacement,
            rrn_replacement=rrn_replacement,
            card_replacement=card_replacement,
            email_replacement=email_replacement,
            bank_account_replacement=bank_account_replacement,
            credit_card_replacement=credit_card_replacement,
            zip_replacement=zip_replacement,
            bitcoin_replacement=bitcoin_replacement,
            url_replacement=url_replacement,
            ip_v6_replacement=ip_v6_replacement,
            ip_v4_replacement=ip_v4_replacement,
        ),
        inputs=text,
        num_workers=num_workers,
    )


def _anonymize(
    text: Union[str, List[str], Tuple[str]],
    phone_number_anonymization: bool = False,
    rrn_anonymization: bool = False,
    card_anonymization: bool = False,
    email_anonymization: bool = False,
    bank_account_anonymization: bool = False,
    credit_card_anonymization: bool = False,
    zip_anonymization: bool = False,
    bitcoin_anonymization: bool = False,
    url_anonymization: bool = False,
    ip_v6_anonymization: bool = False,
    ip_v4_anonymization: bool = False,
    phone_number_replacement: str = "<PHONE_NUMBER>",
    rrn_replacement: str = "<RRN>",
    card_replacement: str = "<CARD>",
    email_replacement: str = "<EMAIL>",
    bank_account_replacement: str = "<BANK_ACCOUNT>",
    credit_card_replacement: str = "<CREDIT_CARD>",
    zip_replacement: str = "<ZIP>",
    bitcoin_replacement: str = "<BITCOIN>",
    url_replacement: str = "<URL>",
    ip_v6_replacement: str = "<IPV6>",
    ip_v4_replacement: str = "<IPV4>",
):
    text = re.sub(PHONE_NUMBER_PATTERN, phone_number_replacement, text) if phone_number_anonymization else text
    text = re.sub(RRN_PATTERN, rrn_replacement, text) if rrn_anonymization else text
    text = re.sub(CARD_PATTERN, card_replacement, text) if card_anonymization else text
    text = re.sub(EMAIL_PATTERN, email_replacement, text) if email_anonymization else text
    text = re.sub(BANK_ACCOUNT_PATTERN, bank_account_replacement, text) if bank_account_anonymization else text
    text = re.sub(CREDIT_CARD_PATTERN, credit_card_replacement, text) if credit_card_anonymization else text
    text = re.sub(ZIP_PATTERN, zip_replacement, text) if zip_anonymization else text
    text = re.sub(BITCOIN_PATTERN, bitcoin_replacement, text) if bitcoin_anonymization else text
    text = re.sub(URL_PATTERN_1, url_replacement, text) if url_anonymization else text
    text = re.sub(URL_PATTERN_2, url_replacement, text) if url_anonymization else text
    text = re.sub(IPV6_PATTERN, ip_v6_replacement, text) if ip_v6_anonymization else text
    text = re.sub(IPV4_PATTERN, ip_v4_replacement, text) if ip_v4_anonymization else text
    return text
